"""Parse single and multiple outputs from bcftools stats into dataframes."""

import argparse
from collections import defaultdict
import os
import re

import pandas as pd
from pandas.api import types as pd_types

from ezcharts.components.ezchart import EZChart
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.base import Snippet
from ezcharts.layout.snippets import DataTable, Tabs
from ezcharts.plots.categorical import barplot


# Categorical types
CATEGORICAL = pd_types.CategoricalDtype(ordered=True)

# Multiple types structure for the different dataframes
coltypes = {
    'SN': {
        'filename': CATEGORICAL,
        'id': CATEGORICAL,
        'MNPs': int,
        'SNPs': int,
        'indels': int,
        'multiallelic SNP sites': int,
        'multiallelic sites': int,
        'no-ALTs': int,
        'others': int,
        'records': int,
        'samples': int,
    },
    'TSTV': {
        'filename': CATEGORICAL,
        'id': CATEGORICAL,
        'ts': int,
        'tv': int,
        'ts/tv': float,
        'ts (1st ALT)': int,
        'tv (1st ALT)': int,
        'ts/tv (1st ALT)': float
    },
    'SiS': {
        'filename': CATEGORICAL,
        'id': CATEGORICAL,
        'allele count': int,
        'number of SNPs': int,
        'number of transitions': int,
        'number of transversions': int,
        'number of indels': int,
        'repeat-consistent': int,
        'repeat-inconsistent': int,
        'not applicable': int
    },
    'AF': {
        'filename': CATEGORICAL,
        'id': CATEGORICAL,
        'allele frequency': float,
        'number of SNPs': int,
        'number of transitions': int,
        'number of transversions': int,
        'number of indels': int,
        'repeat-consistent': int,
        'repeat-inconsistent': int,
        'not applicable': int
    },
    'QUAL': {
        'filename': CATEGORICAL,
        'id': CATEGORICAL,
        'Quality': float,
        'number of SNPs': int,
        'number of transitions (1st ALT)': int,
        'number of transversions (1st ALT)': int,
        'number of indels': int
    },
    'IDD': {
        'filename': CATEGORICAL,
        'id': CATEGORICAL,
        'length (deletions negative)': int,
        'number of sites': int,
        'number of genotypes': int,
        'mean VAF': str   # mean VAF can be a '.' for single-sample VCFs, hence string
    }
}


class BcftoolsStats(Snippet):
    """Generate modified bases summary plots."""

    def __init__(self, theme='epi2melabs', **kwargs):
        """Create bcfstats summary report.

        If multiple single-sample stats are provided, each will be
        displayed in a distinct tab, with the exception of the summary table.

        :param stats: A path to a single-sample bcftools stats file or a pandas
            DataFrame from load_bcfstats
        :param sample_name: A string or list of strings specifying the sample names;
            it has to be of the same length as the input stats files
        """
        super().__init__(styles=None, classes=None)
        raise NotImplementedError(
            "This component is not currently implemented for Bokeh."
        )

        with self:
            # Make table of bcftools stats
            if "stats" in kwargs:
                if isinstance(kwargs["stats"], pd.DataFrame):
                    all_bcfstats = kwargs["stats"]
                else:
                    all_bcfstats = load_bcfstats(
                        kwargs["stats"], sample_names=kwargs["sample_names"])
                # Check that the file is not empty
                if len(all_bcfstats) == 0:
                    raise pd.errors.EmptyDataError('Input stats file is empty')
                # If one sample, save one plot
                for stat, table in all_bcfstats.items():
                    # Add base stats table
                    if stat == 'SN':
                        # we only got a single sample --> no dropdown
                        DataTable.from_pandas(
                            table, export=True, use_index=False)

                    # Showcase indel length
                    if stat == 'IDD':
                        if len(table['id'].unique()) == 1:
                            # we only got a single sample --> no dropdown
                            plt = barplot(
                                data=table,
                                x="length (deletions negative)",
                                y="number of sites")
                            plt._fig.xaxis.major_label_orientation = 90
                            plt._fig.axis_label_text_font_size = 10

                            EZChart(plt, 'epi2melabs')
                        else:
                            # several samples --> use a dropdown menu
                            tabs = Tabs()
                            with tabs.add_dropdown_menu():
                                for fname, df_sample in table.groupby('id'):
                                    with tabs.add_dropdown_tab(fname):
                                        plt = barplot(
                                            data=table,
                                            x="length (deletions negative)",
                                            y="number of sites")
                                        plt._fig.xaxis.major_label_orientation = 90
                                        plt._fig.axis_label_text_font_size = 10
                                        EZChart(plt, 'epi2melabs')


def split_blocks(fname):
    """Split lines of a file into blocks based on comment lines."""
    comment = list()
    data = list()
    state = None
    two_line_comments = ('IDD', 'TSTV', 'SiS', 'AF', 'QUAL', 'ST')
    with open(fname, 'r') as fh:
        lines = fh.readlines()
        for line_num, line in enumerate(lines):
            if line.startswith('#'):
                if state != 'comment' and state is not None:
                    yield comment, data
                    comment, data = list(), list()
                state = 'comment'
                comment.append(line.strip('# ').rstrip())
                # if there are two comments appended and it's a special comment
                # check the next line, if it's another comment, yield the comment
                # and empty data
                if len(comment) == 2 and comment[-1].startswith(two_line_comments):
                    if lines[line_num + 1].startswith('#'):
                        yield comment, []
                        comment, data = list(), list()
            else:
                state = 'data'
                data.append(line.rstrip())
        yield comment, data


def parse_bcftools_stats(fname, sample_name=None):
    """Parse `bcftools stats` output.

    :param fname: file to parse.
    :param sample_name: name of the sample.
    """
    tables = dict()
    filename = fname.split('/')[-1]
    for comment, data in split_blocks(fname):
        fields = [x.rstrip() for x in re.split(r'\[\d+\]', comment[-1])]
        section, fields = fields[0], fields[1:]
        rows = list()
        for d in data:
            items = d.split('\t')
            if items[0] != section:
                raise ValueError("first data field not equal to section key")
            rows.append(items[1:])
        tables[section] = pd.DataFrame(rows, columns=fields)
        # Add file name as initial column in each of them
        tables[section].insert(0, 'filename', filename)
        # Add sample name if provided
        if sample_name:
            tables[section]['id'] = sample_name
        else:
            tables[section]['id'] = filename
    # now some special handling
    SN = tables['SN']
    SN['key'] = SN['key'].apply(
        lambda x: x.replace('number of ', '').rstrip(':'))
    tables['SN'] = pd.DataFrame(
        SN.pivot(index='id', columns='key', values='value').to_records())
    # Add file name as initial column in each of them
    tables['SN'].insert(0, 'filename', filename)
    # Add sample name if provided
    if sample_name:
        tables['SN']['id'] = sample_name
    else:
        tables['SN']['id'] = filename
    # Return tables
    return tables


def load_bcfstats(stats, sample_names=None):
    """Parse multiple bcf stats outputs and combine.

    :param fnames: list of filenames output by `bcftools stats.`
    :param samples_names: list of names of each sample to add to the
        dataframes.
    """
    # Check if it is one or more inputs
    if os.path.isdir(stats):
        filenames = [f"{stats}/{i}" for i in os.listdir(stats)]
    elif os.path.isfile(stats):
        filenames = [stats]
    else:
        raise Exception(f'No valid input: {stats}')

    # Check sample names
    if sample_names is not None:
        if len(sample_names) != len(filenames):
            raise TypeError(
                "`filenames` and `sample_names` should be of equal length.")
        zipped = zip(filenames, sample_names)
        dfs = [
            parse_bcftools_stats(filename, sample_name=sname)
            for (filename, sname) in zipped]
    else:
        dfs = [parse_bcftools_stats(filename) for filename in filenames]

    # Collect every table in one
    all_tables = defaultdict(list)
    for s_tables in dfs:
        # we don't use the 'id' field - its for intersections
        # and unions of two VCFs given to bcftools stats
        for stat, table in s_tables.items():
            all_tables[stat].append(table)
    for stat, tables in all_tables.items():
        if stat in coltypes:
            all_tables[stat] = pd.concat(tables).astype(coltypes[stat])
        else:
            all_tables[stat] = pd.concat(tables).infer_objects()
    return all_tables


def main(args):
    """Entry point to demonstrate the bcftools stats."""
    comp_title = 'BCFtools stats'
    sample_names = args.sample_name if args.sample_name else None
    stats = BcftoolsStats(stats=args.stats, sample_names=sample_names)
    report = ComponentReport(comp_title, stats)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'BCFtools stats',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--stats",
        required=True,
        help="Input bcftools stats file.")
    parser.add_argument(
        "--sample_name",
        help="Name of the sample; valid for a single sample.")
    parser.add_argument(
        "--output",
        default="bcfstats_report.html",
        help="Output HTML file.")
    return parser



"""Get a default report."""
from typing import List, Type

from dominate.tags import a, div, h4, html_tag, p

from ezcharts.components.ezchart import EZChart
from ezcharts.components.reports import Report
from ezcharts.components.theme import (
    EPI2MELabsLogo, LAB_body_resources, LAB_head_resources)
from ezcharts.layout.resource import Resource
from ezcharts.layout.snippets.section import Section


class ComponentReport(Report):
    """A report for displaying a component."""

    headerbar_classes: str = (
        "d-flex align-items-center flex-wrap "
        "justify-content-center bg-dark mb-5")
    headerbar_content_classes: str = (
        "container px-0 d-flex flex-wrap "
        "justify-content-center align-items-center py-3")
    logo_link_classes: str = (
        "d-flex align-items-center pe-5 mb-md-0 me-md-auto "
        "text-decoration-none")
    addendum_classes: str = "py-5 px-4 border-top"
    addendum_content_classes: str = "container px-0"

    def __init__(
        self,
        component_title: str,
        component_element: html_tag,
        logo: Type[html_tag] = EPI2MELabsLogo,
        head_resources: List[Resource] = LAB_head_resources,
        body_resources: List[Resource] = LAB_body_resources,
    ) -> None:
        """Create tag."""
        super().__init__(
            report_title=component_title,
            head_resources=head_resources,
            body_resources=body_resources)

        with self.header:
            with div(className=self.headerbar_classes):
                with div(className=self.headerbar_content_classes):
                    with a(href="/", className=self.logo_link_classes):
                        logo()

        with self.main:
            comp_container = Section('component', component_title)
            comp_container.add(component_element)

        with self.footer:
            with div(className=self.addendum_classes):
                with div(className=self.addendum_content_classes):
                    h4('About this report', className="pb-3")
                    p(
                        "This report was produced using the "
                        f"{component_title} component from ezcharts, "
                        "which is re-usable in other reports."
                    )
                    p(
                        "Oxford Nanopore Technologies products are not "
                        "intended for use for health assessment or to "
                        "diagnose, treat, mitigate, cure or prevent any "
                        "disease or condition.")

    @staticmethod
    def from_plot(plot, path, title="ezChart plot", **kwargs):
        """Create component report from a plot.

        :params path: path to output file.
        :param kwargs: passed to `EZChart`.
        """
        chart = EZChart(plot, "epi2melabs", **kwargs)
        report = ComponentReport(title, chart)
        report.write(path)



"""Get a default report."""
from datetime import datetime
from typing import List, Optional, Type

from dominate.tags import (
    a, button, code, div, h4, html_tag, p, section)

from ezcharts.components.params import ParamsTable
from ezcharts.components.reports import Report
from ezcharts.components.theme import (
    EPI2MELabsLogo, LAB_body_resources, LAB_head_resources)
from ezcharts.components.versions import VersionsTable
from ezcharts.layout.base import IClasses, Snippet
from ezcharts.layout.resource import Resource
from ezcharts.layout.snippets.banner import Banner
from ezcharts.layout.snippets.section import Section
from ezcharts.layout.util import cls


class ILabsAddendumClasses(IClasses):
    """Section html classes."""

    container: str = cls("py-5", "px-0", "border-top")
    inner: str = cls("container", "px-0")


class LabsAddendum(Snippet):
    """A styled footer component for use in a Report."""

    TAG = 'div'

    def __init__(
        self,
        workflow_name: str,
        workflow_version: str,
        classes: ILabsAddendumClasses = ILabsAddendumClasses(),
        use_defaults: bool = True
    ) -> None:
        """Create tag."""
        super().__init__(
            styles=None,
            classes=classes,
            className=classes.container)

        with self:
            self.container = div(className=classes.inner)

        if use_defaults:
            with self.container:
                h4('About this report', className="pb-3")
                p(
                    "This report was produced using the ",
                    code(f"epi2me-labs/{workflow_name}"),
                    f" nextflow workflow ({workflow_version})."
                )
                p(
                    "Oxford Nanopore Technologies products are not "
                    "intended for use for health assessment or to "
                    "diagnose, treat, mitigate, cure or prevent any "
                    "disease or condition.")


class ILabsNavigationClasses(IClasses):
    """Section html classes."""

    spacer: str = cls("d-flex")
    container: str = cls(
        "fixed-top", "d-flex", "align-items-center", "flex-wrap",
        "justify-content-center", "bg-dark")
    inner: str = cls(
        "container", "px-0", "d-flex", "flex-wrap",
        "justify-content-center", "align-items-center", "py-2")
    logo: str = cls(
        "d-flex", "align-items-center", "pe-5", "mb-md-0",
        "me-md-auto", "text-decoration-none")
    dropdown_btn: str = cls("btn", "btn-primary", "dropdown-toggle")
    dropdown_menu: str = cls("dropdown-menu")
    dropdown_item_link: str = cls("dropdown-item")


class LabsNavigation(Snippet):
    """A styled nav component for use in a Report."""

    TAG = 'nav'

    def __init__(
        self,
        logo: Type[html_tag],
        groups: List[str],
        header_height: int = 75,
        classes: ILabsNavigationClasses = ILabsNavigationClasses()
    ) -> None:
        """Create tag."""
        spacer = div(
            className=classes.spacer,
            style=f"margin-top: {header_height}px;")

        super().__init__(
            styles=None,
            classes=classes,
            style=f"min-height: {header_height}px;",
            className=classes.container)

        spacer.add(self)
        with self:
            with div(className=self.classes.inner):
                with a(href="https://labs.epi2me.io/", className=self.classes.logo):
                    logo()

                button(
                    "Jump to section... ",
                    cls=self.classes.dropdown_btn,
                    type="button",
                    id="dropdownMenuButton",
                    data_bs_toggle="dropdown",
                    aria_haspopup="true",
                    aria_expanded="false")

                ngroups = len(groups)
                with div(className=self.classes.dropdown_menu):
                    for count, group in enumerate(groups):
                        setattr(
                            self, group,
                            div(className='', __pretty=False))
                        if count != ngroups - 1:
                            div(cls="dropdown-divider")

    def add_link(
        self,
        group: str,
        link_title: str,
        link_href: str
    ) -> None:
        """Add a header nav link to the header links list."""
        group_list = getattr(self, group)
        with group_list:
            a(
                link_title,
                href=link_href,
                className=self.classes.dropdown_item_link)


class BasicReport(Report):
    """A basic labs-themed report."""

    def __init__(
        self,
        report_title,
        logo: Type[html_tag] = EPI2MELabsLogo,
        head_resources: List[Resource] = LAB_head_resources,
        body_resources: List[Resource] = LAB_body_resources
    ) -> None:
        """Create tag."""
        super().__init__(
            report_title=report_title,
            head_resources=head_resources,
            body_resources=body_resources)
        with self.header:
            self.nav = LabsNavigation(logo=logo, groups=['main', 'meta'])

        with self.main:
            self.intro_content = section(id="intro-content", role="region")
            self.main_content = section(id="main-content", role="region")

    def add_section(
        self,
        title: str,
        link: str,
        overflow: bool = False
    ) -> Section:
        """Add a section to the main_content region."""
        href = self.get_uid('Section')
        self.nav.add_link('main', link, f'#{href}')
        with self.main_content:
            return Section(href, title, overflow=overflow)


class LabsReport(BasicReport):
    """A basic labs-themed report for a workflow."""

    def __init__(
        self,
        report_title,
        workflow_name,
        workflow_params_path: str,
        workflow_versions_path: str,
        workflow_version: str,
        logo: Type[html_tag] = EPI2MELabsLogo,
        head_resources: List[Resource] = LAB_head_resources,
        body_resources: List[Resource] = LAB_body_resources,
        created_date: Optional[str] = None
    ) -> None:
        """Create tag."""
        super().__init__(
            report_title=report_title,
            head_resources=head_resources,
            body_resources=body_resources)

        with self.header:
            self.nav.add_link('meta', 'Versions', '#versions')
            self.nav.add_link('meta', 'Parameters', '#parameters')
            self.intro_content = section(id="intro-content", role="region")
            with self.intro_content:
                self.banner = Banner(report_title, workflow_name)
                self.banner.add_badge("Research use only")
                if not created_date:
                    created_date = datetime.today().strftime('%Y-%m-%d')
                self.banner.add_badge(created_date, bg_class="bg-secondary")
                self.banner.add_badge(workflow_version)

        with self.main:
            self.meta_content = section(id="meta-content", role="region")
            with self.meta_content:
                with Section(
                    "versions",
                    'Software versions',
                    overflow=True
                ):
                    VersionsTable(workflow_versions_path)

                with Section(
                    "parameters",
                    'Workflow parameters',
                    overflow=True
                ):
                    ParamsTable(workflow_params_path)

        with self.footer:
            self.addendum = LabsAddendum(
                workflow_name=workflow_name, workflow_version=workflow_version
            )

    def add_badge(
        self,
        title: str,
        bg_class=None
    ) -> None:
        """Add a badge to the banner."""
        self.banner.add_badge(title, bg_class=bg_class)



"""Get document and helper tags with sensible defaults."""
from dominate.tags import body, head, meta

SCRIPT_KWARGS = dict(__pretty=False, type='text/javascript')
STYLE_KWARGS = dict(__pretty=False)


class DefaultBody(body):
    """Report body element."""

    def __init__(
        self,
        scroll_spy: bool = True,
        **kwargs
    ) -> None:
        """Create tag."""
        if scroll_spy:
            kwargs = {
                "data-bs-spy": "scroll",
                "data-bs-offset": "-150",
                **kwargs
            }
        super().__init__(
            tagname='body', **kwargs)


class DefaultHead(head):
    """Report body element."""

    def __init__(
        self,
        **kwargs
    ) -> None:
        """Create tag."""
        super().__init__(tagname='head', **kwargs)
        with self:
            meta(charset="utf-8")
            meta(
                name="viewport",
                content="width=device-width, initial-scale=1")



"""Re-usable report components."""
from typing import List, Type

from bokeh.embed import components
from dominate.tags import body, footer, head, header, main, style, title
from dominate.util import raw

from ezcharts.components.ezchart import _BokehChart
from ezcharts.layout.base import Snippet
from ezcharts.layout.resource import (
    base_body_resources, base_head_resources, Resource)
from ezcharts.layout.snippets.document import DefaultBody, DefaultHead
from ezcharts.layout.util import write_report


class Report(Snippet):
    """A basic report."""

    TAG: str = 'html'

    def __init__(
        self,
        report_title,
        head_tag: Type[head] = DefaultHead,
        body_tag: Type[body] = DefaultBody,
        head_resources: List[Resource] = base_head_resources,
        body_resources: List[Resource] = base_body_resources
    ) -> None:
        """Create tag."""
        super().__init__(
            styles=None,
            classes=None)

        with self:
            # Adds the generic meta tags for us!
            self.head = head_tag()
            # Enables scroll spy globals for us!
            self.body = body_tag()

        with self.head:
            title(report_title)
            for resource in head_resources:
                resource()

        with self.body:
            self.header = header()
            self.main = main()
            self.footer = footer()
            for resource in body_resources:
                resource()

    def write(self, path):
        """Write a report to file."""
        # check if the report contains `Bokeh` plots
        bokeh_charts = self.get_bokeh_charts()
        if bokeh_charts:
            # get the script + divs for all the plots; then place the div in the
            # corresponding `_BokehChart` div
            bokeh_script, bokeh_divs = components([x.plot._fig for x in bokeh_charts])
            for chart, bokeh_div in zip(bokeh_charts, bokeh_divs):
                with chart:
                    raw(bokeh_div)
            # add the script to the header
            with self.head:
                # make sure the plots fill out the enclosing div
                style(
                    """
                    .bk-Figure {
                        height: 100%;
                        width: 100%;
                    }
                    """
                )
                raw(bokeh_script)

        write_report(path, self)

    def get_bokeh_charts(self):
        """Return all children of the report that are of type `_BokehChart`."""
        bokeh_charts = []

        def _get_charts_in_children(s):
            if not hasattr(s, 'children') or not s.children:
                return
            for child in s.children:
                if isinstance(child, _BokehChart):
                    bokeh_charts.append(child)
                _get_charts_in_children(child)

        _get_charts_in_children(self)
        return bokeh_charts



"""An ezcharts component for common parsers."""
import pandas as pd
from pandas.api import types as pd_types

# Categorical types
CATEGORICAL = pd_types.CategoricalDtype(ordered=True)

# Fix order of modified base type when displying to keep everything consistent
MOD_ORDER = [
    f'{mod}{strand}' for mod in
    ['5mC', '5hmC', '5fC', '5caC', 'modC', '5hmU',
     '5fU', '5caU', 'modT', '6mA', 'modA', '8oxoG',
     'modG', 'Xao', 'modN'] for strand in ['', '_+', '_-']]
# SAM modification codes to human-readable codes
MOD_CONVERT = {
    'm': '5mC',
    'h': '5hmC',
    'f': '5fC',
    'c': '5caC',
    'C': 'modC',
    'g': '5hmU',
    'e': '5fU',
    'b': '5caU',
    'T': 'modT',
    'a': '6mA',
    'A': 'modA',
    'o': '8oxoG',
    'G': 'modG',
    'n': 'Xao',
    'N': 'modN'
}

# Human chromosome order
HSA_CHROMOSOME_ORDER = [f'chr{i}' for i in range(1, 23)] + ['chrX', 'chrY']

# ClinVar/NCBI info
CLINVAR_BASE = "https://www.ncbi.nlm.nih.gov/clinvar/variation/"
NCBI_BASE = "https://www.ncbi.nlm.nih.gov/gene/"
CLINVAR_DOCS_URL = "https://www.ncbi.nlm.nih.gov/clinvar/docs/clinsig/"


# Load faidx
def fasta_idx(faidx, rename=None):
    """Read faidx for the reference fasta."""
    relevant_stats_cols_dtypes = {
        "chrom": CATEGORICAL,
        "length": int,
        "offset1": int,
        "offset2": int,
        "offset3": int,
    }
    try:
        df = pd.read_csv(
            faidx,
            sep="\t",
            names=relevant_stats_cols_dtypes.keys(),
            usecols=relevant_stats_cols_dtypes.keys(),
            dtype=relevant_stats_cols_dtypes,
        )
        if rename:
            df['chrom'] = df['chrom'].cat.rename_categories(rename)
    except pd.errors.EmptyDataError:
        df = pd.DataFrame(columns=relevant_stats_cols_dtypes)
    return df


def make_breaks(minval, maxval, winsize):
    """Create intervals inclusive of last value.

    This takes a initial value, a final value and an interval size.
    It then create the interval using range, and checks if the last value
    matches maxval. If not, it adds it back.
    For example, "range(0, 10001, 1000)" would generate a set of breaks
    [0, 1000, 2000, ..., 10000]
    Leaving out 10001. The function then adds it back to the list.

    This is relevant when creating sliding windows of given size in a
    chromosome to avoid losing the telomeric ends of the genome.
    """
    # Create the intervals
    breaks = list(range(minval, maxval, winsize))
    # Check that the max value is the chr length
    if breaks[-1] < maxval:
        breaks += [maxval]
    return breaks


def add_missing_windows(intervals, faidx, value='value', winsize=25000):
    """Add missing windows to value dataframe.

    If the value refers to a narrow region of the chromosome, this will
    cause to generate a dataframe with either a single-entry, or only
    few entries. This causes the production of poor quality or impossible to
    understand plots.
    This function fixes this by adding the missing leading or trailing
    intervals, matching the full length of the chromosome. Moreover, it will
    check for the presence of gaps in the region, adding back the missing
    points. All added intervals are set to 0-coverage values.
    """
    # Instantiate column types
    relevant_stats_cols_dtypes = {
        "chrom": CATEGORICAL,
        "start": int,
        "end": int,
        value: float,
    }
    # Ensure it is sorted
    intervals = intervals.sort_values(['chrom', 'start'])
    # Get unique chromosomes
    chrs = faidx.query(f'length>={winsize}').chrom.unique().tolist()
    # New intervals
    final_intervals = []
    # Add missing windows for each chromosome
    for chr_id in chrs:
        # Get chromosome entries
        chr_entry = intervals.loc[intervals['chrom'] == chr_id].reset_index(drop=True)
        # First window start
        minval = chr_entry.start.min() if not chr_entry.empty \
            else faidx[faidx['chrom'] == chr_id].length.max()
        # Final window end
        maxval = faidx[faidx['chrom'] == chr_id].length.max()
        # Chromosome length
        chr_len = faidx[faidx['chrom'] == chr_id].length.max()
        # If the minimum value is != 0, add intermediate windows
        if minval != 0:
            # Create the breaks for the intervals
            breaks = make_breaks(0, minval, winsize)
            # Add leading intervals
            final_intervals.append(
                pd.DataFrame(data={
                    'chrom': chr_id,
                    'start': breaks[0:-1],
                    'end': breaks[1:],
                    value: 0}))
        # Append precomputed intervals, checking for breaks
        # in between regions.
        for idx, region in chr_entry.iterrows():
            # To DF
            region = region.to_frame().T
            # Add the first window as default
            if idx == 0:
                final_intervals.append(region)
                continue
            # If the new window start is not the end of the previous,
            # then create new regions
            if region.start.min() != final_intervals[-1].end.max():
                # Create intervals
                breaks = make_breaks(
                    final_intervals[-1].end.max(),
                    region.start.min(),
                    winsize)
                # Add heading    intervals
                final_intervals.append(
                    pd.DataFrame(data={
                        'chrom': chr_id,
                        'start': breaks[0:-1],
                        'end': breaks[1:],
                        value: 0}))
            final_intervals.append(region)
        # If the max value is less than the chromosome length, add these too
        if maxval < chr_len:
            # Create the breaks for the intervals
            breaks = make_breaks(maxval, chr_len, winsize)
            # Create vectors to populate the DF
            chr_vec = [chr_id] * (len(breaks) - 1)
            values = [0] * (len(breaks) - 1)
            # Add tailing intervals
            final_intervals.append(
                pd.DataFrame(data={
                    'chrom': chr_vec,
                    'start': breaks[0:-1],
                    'end': breaks[1:],
                    value: values}))
    return pd.concat(final_intervals).astype(
        relevant_stats_cols_dtypes).reset_index(drop=True)



"""Re-usable external resources."""
from typing import Callable, Optional, Type

from bokeh.resources import INLINE as bk_inline
from dominate.tags import dom_tag, script, style
from dominate.util import raw
from pkg_resources import resource_filename

from ezcharts.layout.util import inline


class Resource:
    """Resource definition."""

    def __init__(
        self,
        path: Optional[str] = None,
        loader: Optional[Callable] = None,
        tag: Optional[Type[dom_tag]] = None,
        func: Optional[Callable] = None,
    ):
        """Construct resource from a path or a function.

        Either `path` or `func` must be given. `path` needs to point to a file in
        `ezcharts/data`, whereas `func` needs to be a callable that returns the resource
        as a string or dominate object.
        """
        # make sure we got only either a path or a function
        if not ((path is None) ^ (func is None)):
            raise ValueError("Resource needs either `path` or `func`.")
        # if we got a path, we require a loader as well
        if path is not None and loader is None:
            raise ValueError("`loader` is required when `path` was provided.")
        self.path = path
        self.loader = loader
        self.tag = tag
        self.func = func

    @property
    def data_file(self):
        """Filepath to resource data file."""
        return (
            resource_filename("ezcharts", f"data/{self.path}")
            if self.path is not None
            else None
        )

    def __call__(self):
        """Render the resource."""
        if self.path is not None:
            loaded = self.loader(self.data_file)
        elif self.func is not None:
            loaded = self.func()
        if self.tag:
            return self.tag(loaded)
        return loaded


def VendorResource(path=None, loader=inline, tag=None, func=None):
    """Fetch a vendor resource."""
    if path is not None:
        path = f"vendor/{path}"
    return Resource(path=path, loader=loader, func=func, tag=tag)


def ScriptResource(path=None, loader=inline, tag=None, func=None):
    """Fetch a script resource."""
    if path is not None:
        path = f"scripts/{path}"
    return Resource(path=path, loader=loader, func=func, tag=tag)


def ImageResource(path=None, loader=inline, tag=None, func=None):
    """Fetch an image resource."""
    if path is not None:
        path = f"images/{path}"
    return Resource(path=path, loader=loader, func=func, tag=tag)


def IconResource(path, loader=inline, tag=None):
    """Fetch an icon resource."""
    return Resource(f'icons/{path}', loader, tag)


def StyleResource(path=None, loader=inline, tag=None, func=None):
    """Fetch a style resource."""
    if path is not None:
        path = f"styles/{path}"
    return Resource(path=path, loader=loader, func=func, tag=tag)


def ThemeResource(path=None, loader=inline, tag=None, func=None):
    """Fetch a theme resource."""
    if path is not None:
        path = f"themes/{path}"
    return Resource(path=path, loader=loader, func=func, tag=tag)


def get_bokeh_js():
    """Get BokehJS library."""
    # make a dict mapping components to indices in `bk_inline.raw_js`
    comp_idx_dict = {comp: i for i, comp in enumerate(bk_inline.components_for('js'))}
    # we only support the basic Bokeh functionality for now (i.e. no widgets etc.)
    return raw(bk_inline.js_raw[comp_idx_dict["bokeh"]])


bokeh_js = ScriptResource(func=get_bokeh_js, tag=script)

echarts_js = VendorResource(
    path='echarts.min.js',
    tag=script,
    loader=inline)

bootstrap_js = VendorResource(
    path='bootstrap-5.3.0/js/bootstrap.bundle.min.js',
    tag=script,
    loader=inline)

datatables_js = VendorResource(
    path='simple-datatables/simple-datatables.js',
    tag=script,
    loader=inline)

datatables_css = VendorResource(
    path='simple-datatables/simple-datatables.css',
    tag=style,
    loader=inline)

base_head_resources = [
    bokeh_js,
    echarts_js,
    datatables_js,
    datatables_css]

base_body_resources = [
    bootstrap_js]



"""Visualizing matrices of data."""
import json

from dominate.tags import div, main, script, style
from dominate.util import raw

from ezcharts.layout.resource import ScriptResource, StyleResource


__all__ = ["metagenomics_sankey"]


def metagenomics_sankey(data):
    """Create a Sankey plot for taxonomic counts.

    :param data (dict): Dictionary with input data.

    Each node in the dict should contain values called `rank`, `count`, and `children`.

    example:
        {"Bacteria": {
            "rank": "superkingdom", "count": 3000, "children": { "Firmicutes": {...
    """
    # get relevant static assets
    sankey_D3_js = ScriptResource("metagenomics-sankey-util.js", tag=script)
    sankey_js = ScriptResource("metagenomics-sankey.js")
    sankey_css = StyleResource("metagenomics-sankey.css", tag=style)

    # add the D3 scripts
    sankey_D3_js()
    # load the actual sankey JS and insert the data
    insert = json.dumps(data).replace('"', '"')
    with open(sankey_js.data_file) as sankey_code_js:
        sankey_js = sankey_code_js.read()
    sankey_data = sankey_js.replace("replace_me", insert.replace('"', '\\"'))

    # create the plot
    with div(className="container"):
        # load sankey style
        sankey_css()

        with main(className="metagenomics-sankey"):
            div(
                raw(
                    """
                    <ul>
                        <li>
                            <label for="sample-select">Select sample</label>
                            <select id="sample-select"></select>
                        </li>
                        <li>
                            <label for="rank-select">Select rank</label>
                            <select id="rank-select"></select>
                        </li>
                        <li>
                            <label for="cutoff-select">Select cutoff</label>
                            <select id="cutoff-select"></select>
                        </li>
                        <li>
                            <button onclick="zoomIn()">[+] Zoom in</button>
                        </li>
                        <li>
                            <button onclick="zoomOut()">[-] Zoom out</button>
                        </li>
                        <li>
                            <button onclick="resetZoom()">[ ] Reset zoom</button>
                        </li>
                    </ul>
                    """
                ),
                id="controls-sankey",
            )
        with div(id="visualisation"):
            # the two divs below will be populated by the script
            div(id="sankey-plot")
            div(id="tooltip")
            script(raw(sankey_data))



"""Make an multiple sequence alignment image."""
import base64
import os

from Bio import AlignIO
from dominate.tags import img, p
from pymsaviz import MsaViz


__all__ = ["msa"]

COLOR_SCHEMES = [
    'Clustal',
    'Zappo',
    'Taylor',
    'Flower',
    'Blossom',
    'Sunset',
    'Ocean',
    'Hydrophobicity',
    'HelixPropensity',
    'StrandPropensity',
    'TurnPropensity',
    'BuriedIndex',
    'Nucleotide',
    'Purine/Pyrimidine',
    'Identity',
    'None'
]


def msa(
        msa_file,
        color='#0ebbb2',
        wrap_lengh=80,
        show_count=True,
        identity_color_min_thr=40,
        show_consensus=False,
        color_scheme="Identity",
        identity=50,
        start=1,
        end=None):
    """
    Make a nice multiple sequence alignment with pymsaviz.

    Args:
        msa_file (str): The path to the MSA file in AFA (Aligned FASTA) format.
        color (str, optional): The color to use for the alignment.
        Defaults to '#0ebbb2'.
        wrap_lengh (int, optional): The maximum length of each line in the
        alignment. Defaults to 80.
        show_count (bool, optional): Whether to show the count of each residue
        in the alignment. Defaults to True.
        show_consensus (bool, optional): Whether to show the consensus
        sequence in the alignment. Defaults to False.
        color_scheme (str, optional): The color scheme to use for the
        alignment. Defaults to "Identity".
        identity (int, optional): The minimum identity threshold for
        highlighting mismatches. Defaults to 50.

    Returns:
        str: The HTML code for displaying the generated alignment image.
    """
    # check MSA file exists
    if not os.path.exists(msa_file):
        return p("Unable to plot MSA. MSA file not found.")

    # check for records in the MSA file
    try:
        alignment = AlignIO.read(msa_file, "fasta")
        if len(alignment) == 0:
            return p("Unable to plot MSA. MSA file does not contain any records.")
    except (ValueError, FileNotFoundError):
        return p("Unable to plot MSA. Invalid FASTA file.")

    # check color scheme is allowed
    if color_scheme not in COLOR_SCHEMES:
        return p(f"Invalid color scheme. Choose from: {', '.join(COLOR_SCHEMES)}")

    mv = MsaViz(
        msa_file,
        wrap_length=wrap_lengh,
        show_count=show_count,
        show_consensus=show_consensus,
        color_scheme=color_scheme,
        start=start,
        end=end)

    pos_ident = []
    ident_list = mv._get_consensus_identity_list()

    # for each position get those with less than dev specified identity. Default is 50.
    for pos, ident in enumerate(ident_list, 1):
        if ident <= identity:
            pos_ident.append(pos)

    # add dots over mismatches
    mv.add_markers(
        pos_ident, marker="o", color="#c22932")

    # change to ONT colours
    mv.set_plot_params(
        identity_color=color, identity_color_min_thr=identity_color_min_thr)

    mv.savefig("msa.png")

    with open("msa.png", "rb") as image_file:
        data = base64.b64encode(image_file.read())

    return img(
        src=f"data:image/png;base64,{data.decode('utf-8')}",
        cls="img-fluid")



"""Entrypoint to demo an ezCharts plot."""
import argparse

from pkg_resources import resource_filename

from ezcharts.components.reports.comp import ComponentReport
from ezcharts.plots import Plot


def main(args):
    """Entry point to demonstrate an ezChart."""
    plt = Plot.parse_file(args.plot_spec)
    ComponentReport.from_plot(plt, args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'ezChart Demo',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--plot_spec",
        default=resource_filename('ezcharts', "data/test/plot-spec.json"),
        help=(
            "A JSON file defining an eCharts plot",
            "key/values."
        ))
    parser.add_argument(
        "--output",
        default="ezcharts_plot_report.html",
        help="Output HTML file.")
    return parser



"""Categorical plots."""
from collections import Counter

from bokeh.models import ColumnDataSource, FactorRange, HoverTool, Whisker
import bokeh.transform as bkt
import numpy as np
import pandas as pd
from seaborn.categorical import _BarPlotter

from ezcharts.plots import BokehPlot


__all__ = [
    "catplot", "stripplot", "swarmplot", "boxplot", "violinplot",
    "boxenplot", "pointplot", "barplot", "countplot"]


def catplot(*args, **kwargs):
    """Figure-level interface for drawing categorical plots in a grid."""
    raise NotImplementedError


def stripplot(*args, **kwargs):
    """Draw a scatterplot where one variable is categorical."""
    raise NotImplementedError


def swarmplot(*args, **kwargs):
    """Draw a categorical scatterplot with non-overlapping points."""
    raise NotImplementedError


def boxplot(
        data=None, *, x=None, y=None, hue=None, order=None, hue_order=None,
        orient=None, color=None, palette=None, saturation=1, fill=True,
        dodge=None, width=None, gap=None, whis=1.5, linecolor='auto', linewidth=1,
        fliersize=6, hue_norm=None, native_scale=None, log_scale=None,
        formatter=None, legend=None, ax=None, **kwargs):
    """Draw a box plot to show distributions with respect to categories."""
    # deal with stuff we haven't implemented, yet
    not_implemented = [
        hue, hue_order, orient, dodge, width, ax, formatter, legend,
        gap, hue_norm, native_scale, log_scale]
    for i in not_implemented:
        if i is not None:
            raise NotImplementedError(
                f"The parameter with the value {i} is not yet implemented.")

    # use our default palette if no colour options were provided
    if palette is None and color is None:
        palette = BokehPlot.colors

    # we are going to group by our x
    groups = data[x].unique()

    # compute quantiles
    qs = data.groupby(x)[y].quantile([0.25, 0.5, 0.75])
    qs = qs.unstack().reset_index()
    qs.columns = [x, "q1", "q2", "q3"]
    df = pd.merge(data, qs, on=x, how="left")

    # if the user specifies an order for their categorical variables
    if isinstance(order, list):
        # check to make sure user has included all categorical variables
        check = Counter(order) == Counter(groups)
        if check:
            groups = order
        else:
            raise ValueError(
                f"all categories ({groups}) not present in order ({order})")

    # compute IQR outlier bounds
    iqr = df.q3 - df.q1

    # whiskers
    if isinstance(whis, float):
        df["upper"] = df.q3 + whis * iqr
        df["lower"] = df.q1 - whis * iqr
    else:
        raise NotImplementedError(
            f"'whis' must be float, {type(whis)} is not supported or implemented")

    source = ColumnDataSource(df)

    # quantile boxes color
    if not color:
        color = bkt.factor_cmap(x, palette, groups)

    # use seaborn bits and bobs for color
    if not fill:
        color = None
        linecolor = "black"
    else:
        if linecolor == 'auto':
            linecolor = "black"

    # outlier range
    whisker = Whisker(
        base=x, upper="upper", lower="lower", source=source,
        line_width=linewidth, line_color=linecolor)

    whisker.upper_head.size = whisker.lower_head.size = 20

    # outliers
    outliers = df[~df[y].between(df.lower, df.upper)]

    # we need to accpount for whskers and outliers in the y-range
    y_min = df["lower"].min()
    y_max = df["upper"].max()

    # if there are outliers then we need to take those into account
    if not outliers.empty:
        y_min = np.min([outliers[y].min(), df["lower"].min()])
        y_max = np.max([outliers[y].max(), df["upper"].max()])

    y_range = (1.1 * y_min, 1.1 * y_max)

    plt = BokehPlot(x_range=groups, y_range=y_range)
    p = plt._fig
    p.add_layout(whisker)

    p.vbar(
        x, 0.7, "q2", "q3", source=source, fill_color=color, line_color=linecolor,
        line_width=linewidth, fill_alpha=saturation)

    p.vbar(
        x, 0.7, "q1", "q2", source=source, fill_color=color, line_color=linecolor,
        line_width=linewidth, fill_alpha=saturation)

    p.scatter(x, y, source=outliers, size=fliersize, color="black", alpha=0.5)

    p.xgrid.grid_line_color = None
    p.xaxis.axis_label = x.capitalize()
    p.yaxis.axis_label = y.capitalize()

    # Add tooltips
    hover = plt._fig.select(dict(type=HoverTool))
    hover.tooltips = [
        ("Value", "@value"),
    ]

    return plt


def violinplot(*args, **kwargs):
    """Draw a combination of boxplot and kernel density estimate."""
    raise NotImplementedError


def boxenplot(*args, **kwargs):
    """Draw an enhanced box plot for larger datasets."""
    raise NotImplementedError


def pointplot(*args, **kwargs):
    """Show point estimates and confidence intervals using scatter plot."""
    # NOTE: default estimator is mean
    raise NotImplementedError


def barplot(
    data=None, *, x=None, y=None, hue=None, order=None, hue_order=None,
    estimator='mean', errorbar=('ci', 95), n_boot=1000, units=None, seed=None,
    orient=None, color=None, palette=None, saturation=1.0, width=0.8,
    errcolor='.26', errwidth=None, capsize=None, dodge=True, ci='deprecated',
    ax=None, nested_x=False, **kwargs,
):
    """Show point estimates as rectangular bars.

    Contrary to the seaborn implementation, setting `dodge=False` does not
    result in overlaying the bars, but rather stacking them.

    nested_x: create a nested plot instead of a grouped plot, which has two X
    axis grouping (hue as outer group)
    """
    # use our default palette if no colour options were provided
    if palette is None and color is None:
        palette = BokehPlot.colors

    # A plot can either be nested or stacked, not both
    if nested_x:
        if hue is None or not dodge:
            raise ValueError(
                "`hue` and `dodge` need to be set when passing `nested_x=True`"
            )
        if orient == "h":
            raise ValueError(
                "`nested_x=True` can only work with `orient='v'`"
            )

    # Create bar plot with seaborn
    plotter = _BarPlotter(
        x,
        y,
        hue,
        data,
        order,
        hue_order,
        estimator,
        errorbar,
        n_boot,
        units,
        seed,
        orient,
        color,
        palette,
        saturation,
        width,
        errcolor,
        errwidth,
        capsize,
        dodge,
    )

    # Nested X labeling requires the factors to be provided as a list of
    # tuples with the hue and x levels for each value in the dataframe:
    # factors = [
    #   ("hue1", "x1"),
    #   ("hue1", "x2"),
    #   ("hue2", "x1"),
    #   ("hue2", "x2"),
    # ]
    if nested_x:
        factors = [
            (str(x0), str(x1)) for x0 in plotter.hue_names for x1 in plotter.group_names
        ]
        group_names = FactorRange(*factors)
    else:
        group_names = [str(x) for x in plotter.group_names]

    if plotter.orient == "v":
        plt = BokehPlot(x_range=group_names)
    else:
        plt = BokehPlot(y_range=group_names)

    p = plt._fig

    # Define plot orientation
    # If both hue and dodge=False are provided, make a stacked bar chart
    if plotter.orient == "v":
        plot_bars_func = p.vbar_stack if not dodge and hue else p.vbar
    else:
        plot_bars_func = p.hbar_stack if not dodge and hue else p.hbar

    if plotter.plot_hues is None:
        # simple barplot (i.e. only a single group of bars)
        if plotter.orient == "v":
            plot_bars_func(
                x=group_names,
                top=plotter.statistic,
                width=0.9,
                color=plotter.colors.as_hex(),
                **kwargs
            )
        else:
            plot_bars_func(
                y=group_names,
                right=plotter.statistic,
                height=0.9,
                color=plotter.colors.as_hex(),
                **kwargs
            )
    elif nested_x:
        # Nested X uses different labelling of the input values
        # x: list of factors described above
        # y: values in a single list/vector
        # legend_name: point to to the legend labelling
        # Create the legend labels
        dual_legend_label = [
            str(x0) for x0 in plotter.hue_names for _ in plotter.group_names
        ]
        # Create the data structure
        data = dict(
            x=factors, y=np.hstack(plotter.statistic.T), legend_name=dual_legend_label
        )
        # Convert to ColumnDataSource
        data = ColumnDataSource(data)
        plot_bars_func(
            x="x",
            top="y",
            source=data,
            line_color="white",
            legend_field="legend_name",
            color=bkt.factor_cmap(
                "x", palette=plotter.colors.as_hex(), factors=plotter.hue_names, end=1
            ),
            **kwargs,
        )
    else:
        data = dict(groups=plotter.group_names)
        data.update(dict(zip(plotter.hue_names, plotter.statistic.T)))
        data = ColumnDataSource(data)
        if not dodge:
            # stacked bars (define orientation-specific kwargs first)
            if plotter.orient == "v":
                extra_kwargs = dict(x="groups", width=0.95)
            else:
                extra_kwargs = dict(y="groups", height=0.95)
            plot_bars_func(
                plotter.hue_names,
                color=plotter.colors.as_hex(),
                source=data,
                legend_label=plotter.hue_names,
                **extra_kwargs,
                **kwargs,
            )
        else:
            # grouped bars (we need to add a series for each hue level)
            for hue_level, offset, color in zip(
                plotter.hue_names,
                plotter.hue_offsets,
                plotter.colors.as_hex(),
            ):
                # define orientation-specific kwargs
                if plotter.orient == "v":
                    extra_kwargs = dict(
                        x=bkt.dodge("groups", offset, range=p.x_range),
                        top=hue_level,
                        width=plotter.nested_width,
                    )
                else:
                    extra_kwargs = dict(
                        y=bkt.dodge("groups", offset, range=p.y_range),
                        right=hue_level,
                        height=plotter.nested_width,
                    )
                plot_bars_func(
                    source=data,
                    color=color,
                    line_color="white",
                    legend_label=hue_level,
                    **extra_kwargs,
                    **kwargs,
                )

        p.legend.orientation = "horizontal"
        # even though we call `add_layout()` to pull the legend outside of the plot area
        # below, we still need to set `location` to "top" here to make sure the legend
        # is centered properly
        p.legend.location = "top"
        p.add_layout(p.legend[0], "above")

    if plotter.orient == "v":
        p.xgrid.grid_line_color = None
        p.y_range.start = 0
    else:
        p.ygrid.grid_line_color = None
        p.x_range.start = 0

    if not nested_x:
        p.xaxis.axis_label = x.capitalize()
        p.yaxis.axis_label = y.capitalize()

    plt_df = plt._fig.renderers[0].data_source.to_df()
    hover = plt._fig.select(dict(type=HoverTool))
    if 'top' in plt_df.columns:
        hover.tooltips = [(y, "@top")]
    elif not nested_x:
        hover.tooltips = [
            (colname, '@{' + colname + '}')
            for colname in plt_df.columns
            if colname != 'groups'
        ]
    else:
        hover.tooltips = [
            (y, '@y')
        ]
    return plt


def countplot(*args, **kwargs):
    """Show the counts of observations in each categorical bin using bars."""
    raise NotImplementedError



"""An ezcharts component for loading ClinVar annotated VCF files."""
import argparse
import os

from dominate.tags import a, span
from natsort import natsort_keygen
import pandas as pd
from pandas.api import types as pd_types
from pysam import VariantFile

from ezcharts.components.common import CLINVAR_BASE, NCBI_BASE
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.base import Snippet
from ezcharts.layout.snippets import DataTable, Tabs
from ezcharts.layout.util import isolate_context

# Categorical types
CATEGORICAL = pd_types.CategoricalDtype(ordered=True)


class ClinVarTable(Snippet):
    """Generate modified bases summary plots."""

    def __init__(self, theme='epi2melabs', **kwargs):
        """Create ClinVar data table.

        If multiple single-sample VCFs are provided, each will be
        displayed in a distinct tab.

        :param vcf_fn: A path to a single-sample vcf file annotated with ClinVar fields
            or a pandas DataFrame from load_clinvar_vcf
        :param benign: A boolean specifying whether to include the benign sites in
            the report
        :param all_sites: A boolean specifying whether to include the every sites in
            the report
        """
        super().__init__(styles=None, classes=None)

        with self:
            # Make table of clinvar sites
            if "vcf_fn" in kwargs:
                if isinstance(kwargs["vcf_fn"], pd.DataFrame):
                    clinvar_df = kwargs["vcf_fn"]
                else:
                    clinvar_df = load_clinvar_vcf(
                        kwargs["vcf_fn"],
                        benign=kwargs["benign"],
                        all_sites=kwargs["all_sites"])
                # Check that the file is not empty
                if clinvar_df.empty:
                    raise pd.errors.EmptyDataError('Input VCF is empty')
                # If one sample, save one plot
                if len(clinvar_df['Sample'].unique()) == 1:
                    # we only got a single sample --> no dropdown
                    DataTable.from_pandas(
                        clinvar_df, export=True, use_index=False)
                else:
                    # several samples --> use a dropdown menu
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_id, df_sample in clinvar_df.groupby('sample_name'):
                            with tabs.add_dropdown_tab(sample_id):
                                DataTable.from_pandas(
                                    df_sample, export=True, use_index=False)


# Function to load a single VCF file.
def load_vcf(vcf_fn, benign=False, all_sites=False):
    """Import a single vcf file."""
    # Start processing the data
    data = []
    columns = {
        'Sample': CATEGORICAL,
        "Chrom": CATEGORICAL,
        "Pos": int,
        "Gene(s)": CATEGORICAL,
        "ClinVar": CATEGORICAL,
        "Significance": CATEGORICAL,
        "Type": CATEGORICAL,
        "Consequence": CATEGORICAL,
        "HGVSc": CATEGORICAL,
        "HGVSp": CATEGORICAL,
    }

    # Load the VCF. If empty, return empty DF.
    try:
        vcf_file = VariantFile(vcf_fn)
    except ValueError:
        return pd.DataFrame(columns=columns).astype(columns)

    # Get sample name
    if len(vcf_file.header.samples) != 1:
        raise ValueError(f'File {vcf_fn} has more than one sample')
    else:
        sample_name = vcf_file.header.samples[0]

    # Process each record in the VCF file
    for variant in vcf_file.fetch():
        # First thing first, we check if the site has an alternative allele
        # and if not we skip it.
        if not variant.alts:
            continue

        # clinvar significance
        if 'CLNSIG' not in variant.info:
            # if absent and benign are requested, then set to "Not provided"
            if all_sites:
                significance = 'Not provided'
            # otherwise skip
            else:
                continue
        # if present, process appropriately
        else:
            significance = ", ".join(variant.info['CLNSIG'])

        # If the site is benign, ignore it (unless otherwise specified)
        if 'benign' in significance.lower() and not benign and not all_sites:
            continue
        else:
            # NCBI gene URL
            try:
                all_ncbi_urls = []
                clinvar_gene_string = variant.info['GENEINFO']
                # CW-2783: If there is a tuple, each element will be split
                # independently and added to the list. Duplicated entries are
                # dropped using a tuple/set/tuple conversion.
                if type(clinvar_gene_string) is tuple:
                    all_genes = list(
                        set([
                            i for string in clinvar_gene_string
                            for i in string.split('|')
                        ]))
                else:
                    all_genes = clinvar_gene_string.split('|')

                for gene in all_genes:
                    gene_symbol, gene_id = gene.split(':')
                    with isolate_context():
                        ncbi_url = str(a(
                            gene_symbol,
                            href=f'{NCBI_BASE}{gene_id}'))
                        all_ncbi_urls.append(ncbi_url)
                ncbi_gene_url = ", ".join(all_ncbi_urls)
            except KeyError:
                ncbi_gene_url = "No affected genes found"

            # multiple ClinVar IDs possible, separated by ';'
            clinvar_url_list = []
            clinvar_id = variant.id
            if not clinvar_id or clinvar_id == '.':
                clinvar_urls_for_report = '.'
            else:
                all_clinvar_ids = clinvar_id.split(';')
                for each_id in all_clinvar_ids:
                    with isolate_context():
                        clinvar_url = str(a(
                            each_id, href=f'{CLINVAR_BASE}{each_id}'))
                        clinvar_url_list.append(clinvar_url)
                clinvar_urls_for_report = ", ".join(clinvar_url_list)

            # tidy up significances
            significance = significance.replace("_", " ").replace(
                "|", ", ").capitalize()

            # tidy up variant types
            try:
                variant_type = variant.info['CLNVC']
                # CW-2783: There should be just one variant type, but in case multiple
                # alleles are present the variant types are combined in a
                # comma-separated list (duplicates are still removed via set).
                if type(variant_type) is tuple:
                    variant_type = [
                        vt.replace("_", " ").capitalize() for vt in set(variant_type)
                    ]
                    variant_type = ','.join([
                        'SNV' if v == 'Single nucleotide variant' else v
                        for v in variant_type
                    ])
                else:
                    variant_type = variant_type.replace("_", " ").capitalize()
                    if variant_type == 'Single nucleotide variant':
                        variant_type = 'SNV'
            except KeyError:
                # If not CLNVC, use variant type from pysam
                variant_type = variant.alleles_variant_types[1]

            # tidy up consequences
            consequences = []
            try:
                all_consequences = variant.info['MC']
                for each_conseq in all_consequences:
                    ontology, consequence = each_conseq.split('|')
                    consequence = consequence.replace("_", " ").capitalize()
                    consequence = consequence.replace(
                        " prime utr", "' UTR")
                    consequences.append(consequence)
                consequences = ", ".join(consequences)
            except KeyError:
                consequences = "No consequences found"

            # HGVS p. and c. come from the SnpEff ANN field, we will take the first
            # available 'NM_' number
            hgvsc = ""
            hgvsp = ""
            all_variant_info = variant.info['ANN']
            for each_variant in all_variant_info:
                fields = each_variant.split('|')
                # find the first NM (because there are some XM records present)
                if fields[6].startswith("NM"):
                    hgvsc = f"{fields[6]}:{fields[9]}"
                    hgvsp = fields[10] if fields[10] != "" else "-"
                    break
                # if there are only XMs then display the first one of those instead
                else:
                    hgvsc = f"{fields[6]}:{fields[9]}"
                    hgvsp = fields[10] if fields[10] != "" else "-"

        # Create record tuple
        record = (
            sample_name, variant.chrom, variant.pos, ncbi_gene_url,
            clinvar_urls_for_report, significance, variant_type,
            consequences, hgvsc, hgvsp
        )
        data.append(record)

    # Put together the dataframe
    df = pd.DataFrame(data, columns=columns)

    # Check if it is empty, and if so return a "No sites" line for the sample
    if df.empty:
        record = (
            sample_name, 'No sites', '', '', '',
            '', '', '', '', ''
        )
        return pd.DataFrame(data, columns=columns).astype(columns)

    # get unique list of all significances in the df
    sample_significances = df['Significance'].unique()

    # desired order of ClinVar significances
    significance_order = [
        "Pathogenic", "Pathogenic, low penetrance",
        "Likely pathogenic", "Likely pathogenic, low penetrance",
        "Uncertain significance", "Conflicting interpretations of pathogenicity",
        "Risk factor", "Established risk allele", "Likely risk allele",
        "Uncertain risk allele", "Association", "Confers sensitivity", "Affects",
        "Protective", "Association not found", "Drug response", "Other",
        "Not provided", "Benign"]

    """re-order variant significances found in the sample to the desired order
    also ensures there are no duplicate significances as this will cause a
    problem when setting 'Significance' to a categorical column later
    """
    reordered = [
        y for x in significance_order for y in sample_significances if y.startswith(x)]
    reordered_unique = sorted(set(reordered), key=reordered.index)

    # define custom sort for chromosomes
    def nat_sort_chromosome(col):
        if col.name == "Chrom":
            return col.apply(natsort_keygen())
        return col

    # sort the sample ClinVar results using the re-ordered significances
    df["Significance"] = pd.Categorical(
        df["Significance"], categories=reordered_unique)
    df.sort_values(
        by=["Significance", "Chrom", "Pos"], inplace=True, key=nat_sort_chromosome)

    return df.astype(columns)


# Load mod bedMethyl file
def load_clinvar_vcf(vcf_fn, benign=False, all_sites=False):
    """Convert ClinVar VCF to sorted df."""
    if os.path.isdir(vcf_fn):
        dfs = [
            load_vcf(f"{vcf_fn}/{i}", benign=benign, all_sites=all_sites)
            for i in os.listdir(vcf_fn)
            if i.endswith(('.vcf', '.vcf.gz', 'bcf'))]
    elif os.path.isfile(vcf_fn) and vcf_fn.endswith(('.vcf', '.vcf.gz', 'bcf')):
        dfs = [load_vcf(vcf_fn, benign=benign, all_sites=all_sites)]
    else:
        raise Exception(f'No valid input: {vcf_fn}')

    return pd.concat(dfs)


def format_clinvar_table(df):
    """Format ClinVar dataframe to add badges where required."""
    df_for_table = df.copy()
    # uncategorise the 'Significance' column so we can add badges
    df_for_table['Significance'] = df_for_table['Significance'].astype(
        df_for_table['Significance'].cat.categories.to_numpy().dtype)

    significance_badges = {
        "Pathogenic": "badge bg-danger",
        "Likely pathogenic": "badge bg-warning",
        "Uncertain": "badge bg-primary"
    }
    for idx, sig in df["Significance"].items():
        for significance, class_name in significance_badges.items():
            if sig.startswith(significance):
                df_for_table.loc[idx, "Significance"] = str(span(sig, cls=class_name))
                break
    return (df_for_table)


def main(args):
    """Entry point to demonstrate a ClinVar data table."""
    comp_title = 'ClinVar data'
    seq_sum = ClinVarTable(
        vcf_fn=args.vcf,
        benign=args.benign,
        all_sites=args.all_sites
        )
    report = ComponentReport(comp_title, seq_sum)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'ClinVar summary',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--vcf",
        default=None,
        required=True,
        help="Input VCF file.")
    parser.add_argument(
        "--benign",
        action='store_true',
        help="Report benign sites in the VCF file.")
    parser.add_argument(
        "--all_sites",
        action='store_true',
        help="Report all sites in the VCF file.")
    parser.add_argument(
        "--output",
        default="clinvar_report.html",
        help="Output HTML file.")
    return parser



"""Visualizing matrices of data."""
import numpy as np
import pandas as pd
import seaborn as sns

import ezcharts as ezc
from ezcharts.components.common import make_breaks


__all__ = ["karyomap"]


def karyomap(
        df, chrom, pos, value, stats='count',
        order=None, palette="crest", window_size=1000000,
        ref_lengths=None, borders=True, labels=True,
        min_sequence_size=None, include_small_ctgs=False):
    """
    Plot chromosomal heatmap in windows of a given size.

    Accepts a dataset with at least three columns:
    - chromosome (y axis)
    - positions (x axis)
    - values (colour intensity)

    The dataset is collected in bins of given size, with the following stats
    availabe:
    - count
    - mean
    - median

    A custom ordering can be specified, but if not it will sort by sequence size.

    By default, the plot will represent only sequences longer than the window size
    (i.e. if window_size=1Mb, then sequences of 999,999bp and smaller won't be shown).

    The default palette uses dark-to-light colours for lower-to-higher values.

    If ref_lengths is not provided, the script will use the highest position in the
    dataset.
    """
    if not isinstance(df, pd.DataFrame):
        df = pd.DataFrame(df)

    def build_intervals(ctgid, start, end, size):
        """Build a dataframe of intervals."""
        breaks = make_breaks(start, end, size)
        return pd.DataFrame(
            data={
                'chr': ctgid,
                'start': breaks[0:-1],
                'end': breaks[1:]})

    # Compute central value.
    def compute_value(data, x, col='values', stats=stats):
        """Return the average of a given column in a region."""
        data_ = data[(data[chrom] == x.chr) &
                     (data[pos] >= x.start) &
                     (data[pos] < x.end)]
        if stats == 'count':
            return data_.shape[0]
        if stats == 'median':
            return data_[col].median()
        elif stats == 'mean':
            return data_[col].mean()
        else:
            raise ValueError("stats must be one of 'count', 'mean', or 'median'")

    # If no ref_length, define it from the positions
    if ref_lengths is None:
        ref_lengths = df.groupby(chrom).max()[pos].reset_index()
        ref_lengths.columns = ['chrom', 'length']

    # Check that ref_length contains the right columns.
    if 'length' not in ref_lengths.columns:
        raise ValueError(
            'No "length" column found.',
            'Import the fai with fasta_idx() and try again.')
    if 'chrom' not in ref_lengths.columns:
        raise ValueError(
            'No "chrom" column found.',
            'Import the fai with fasta_idx() and try again.')

    # Drop all sequences that are smaller than the window size.
    # This prevents having N rows in the heatmap, where N is >2000 for hg38
    # The default behaviour is to keep sequences larger than the window size,
    # but alternatively a specific size can be enforced with min_sequence_size.
    # Finally, users can use all contigs regardless with include_small_ctgs
    if include_small_ctgs is False:
        if isinstance(min_sequence_size, int):
            ref_lengths = ref_lengths.loc[ref_lengths.length > min_sequence_size]
        else:
            ref_lengths = ref_lengths.loc[ref_lengths.length > window_size]

    # Build intervals to visualize
    intervals = pd.concat([
        build_intervals(r.chrom, 0, r.length, window_size)
        for idx, r in ref_lengths.iterrows()])

    # Compute median -log10(P-value) for each interval
    intervals['value'] = intervals.apply(
        lambda x: compute_value(df, x, col=value, stats=stats), axis=1)
    intervals = intervals.fillna(0).replace([np.inf, -np.inf], 0)

    # Transpose the matrix
    mat = intervals.pivot(index='start', columns='chr', values='value')

    # Define new index as window intervals with format start_pos-end_pos
    new_idx = pd.Series(
        [f'{c}-{c + window_size}'
            for c in mat.index.astype('int')])
    mat = mat.set_index(new_idx)

    # Keep chromosomes only and reorder them.
    # Reverse ordering needed to have the first on top.
    if order:
        cols = [
            col for col in order if col in mat.columns][::-1]
    else:
        cols = ref_lengths.sort_values('length', ascending=True).chrom.values
    mat = mat[cols]

    # Define color map
    colmap = sns.color_palette(palette, n_colors=1000, as_cmap=False).as_hex()

    # Create the heatmap
    plt = ezc.heatmap(mat, annot=False, cmap=colmap)
    # Make Y-axis labels smaller and show all of them
    plt.yAxis.axisLabel = dict(fontSize=10, interval=0)
    # Remove unnecessary X labels since too many
    plt.xAxis.axisLabel = dict(show=False)
    # Add borders to the heatmap cells
    if borders:
        for s in plt.series:
            s.itemStyle = dict(
                borderColor='white',
                borderWidth=1
            )
    # Add tooltips
    if labels:
        plt.tooltip = dict(show=True)
    return plt



"""Re-usable external resources."""
from dominate.tags import div, script, style

from ezcharts.components.ezchart import EZChartTheme
from ezcharts.layout.resource import (
    base_body_resources, base_head_resources,
    ImageResource, ScriptResource, StyleResource, ThemeResource)
from ezcharts.layout.util import inline, load_json, transpile


class EPI2MELabsLogo(div):
    """Labs logo element."""

    def __init__(self) -> None:
        """Create a div with an SVG logo inside."""
        super().__init__(
            inline(ImageResource('LAB_logo.svg').data_file),
            tagname='div',
            style="width: 35px; height: 35px;",
            className="d-flex",
            alt="EPI2ME Labs Logo")


LAB_CSS = StyleResource(
    path='epi2melabs.scss',
    tag=style,
    loader=transpile)

LAB_JS = ScriptResource(
    path='epi2melabs.js',
    tag=script,
    loader=inline)

LAB_CHART_THEME = ThemeResource(
    path='epi2melabs.json',
    tag=EZChartTheme,
    loader=load_json)


LAB_head_resources = [LAB_CSS, *base_head_resources, LAB_CHART_THEME]
LAB_body_resources = [LAB_JS, *base_body_resources]


class ONDLogo(div):
    """OND logo element."""

    def __init__(self) -> None:
        """Create an img with the data URI logo."""
        super().__init__(
            inline(ImageResource('OND_logo.svg').data_file),
            tagname='div',
            style="width: 35px; height: 35px;",
            className="d-flex",
            alt="OND Logo")


# TODO: Create
OND_CSS = StyleResource(
    path='ond.scss',
    tag=style,
    loader=transpile)

OND_JS = ScriptResource(
    path='ond.js',
    tag=script,
    loader=inline)


OND_CHART_THEME = ThemeResource(
    path='ond.json',
    tag=EZChartTheme,
    loader=load_json)

OND_head_resources = [OND_CSS, *base_head_resources, OND_CHART_THEME]
OND_body_resources = [OND_JS, *base_body_resources]


class ONTLogo(div):
    """ONT logo element."""

    def __init__(self) -> None:
        """Create an img with the data URI logo."""
        super().__init__(
            inline(ImageResource('OND_logo.svg').data_file),
            tagname='div',
            style="width: 35px; height: 35px;",
            className="d-flex",
            alt="ONT Logo")


# TODO: Create
ONT_CSS = StyleResource(
    path='ont.scss',
    tag=style,
    loader=transpile)

ONT_JS = ScriptResource(
    path='ont.js',
    tag=script,
    loader=inline)

ONT_head_resources = [ONT_CSS, *base_head_resources]
ONT_body_resources = [ONT_JS, *base_body_resources]



"""Get default table layouts."""
from collections.abc import Iterable
from typing import List, Optional, Union

from dominate.tags import button, script, table, tbody, td, th, thead, tr
from dominate.util import raw
import pandas as pd

from ezcharts.layout.base import IClasses, Snippet
from ezcharts.layout.util import cls, render_template


class ITableClasses(IClasses):
    """Table html classes."""

    container: str = cls("table-responsive")
    table: str = cls("table", "table-striped", "table-hover")


class DataTable(Snippet):
    """A styled datatable wrapped in a div."""

    TAG = 'div'

    def __init__(
        self,
        headers: Optional[List[str]] = None,
        page_length: Optional[int] = 10,
        paging: Optional[bool] = True,
        searchable: Optional[bool] = True,
        file_name: Optional[str] = 'datatable-export',
        export: Optional[bool] = False,
        classes: ITableClasses = ITableClasses(),
    ) -> None:
        """
        Create table.

        The default constructor should only be used in special cases. Use the
        alternative constructors `.from_pandas()` or `.from_dict()` instead.

        :param headers: List of strings for headers or multiple lists for multilevel
            headers. If `None`, no header is produced, defaults to `None`
        :param page_length: Page length of the table, defaults to `10`
        :param paging: Whether to allow paging, defaults to `True`
        :param classes: HTML classes, defaults to `ITableClasses()`
        """
        super().__init__(
            styles=None,
            classes=classes,
            className=classes.container)

        with self:
            with table(
                id=self.uid + '_inner',
                className=self.classes.table
            ):
                self.head = thead()
                if headers is not None:
                    with self.head:
                        # check for multi-level headers (i.e. if `headers` is a list of
                        # lists)
                        if not self._is_multilevel_header(headers):
                            # single-level header --> wrap in list
                            headers = [headers]
                        for header_level in headers:
                            with tr():
                                for _header in header_level:
                                    th(_header, scope="col")
                self.body = tbody()

            # Prepare table
            datatable_render = """
                var {{ id }}_table = new simpleDatatables.DataTable( \
                    '#{{ id }}_inner', { \
                    searchable: {{ searchable }}, \
                    pageLength: {{ page_length }}, \
                    paging: {{ paging }} \
                })"""

            if export:
                datatable_render, exportCSVButton = self._export_table(datatable_render)
            else:
                exportCSVButton = None

            script(raw(render_template(
                datatable_render,
                id=self.uid, paging=str(paging).lower(),
                searchable=str(searchable).lower(), page_length=page_length,
                button_id=exportCSVButton, file_name=file_name)))

    def _is_multilevel_header(self, headers: list) -> bool:
        """Check if header list is a multi-level header (i.e. a list of lists).

        :param headers: List of strings for headers or multiple lists for multilevel
            headers.
        :raises ValueError: Raise error when `headers` contains iterables as well as
            non-iterables (e.g. lists and also strings)
        :return: `bool` whether the header is multilevel
        """
        # we got a multilevel header if all items in `headers` are non-string iterables
        # (lists, tuples, etc.)
        headers_iterables = [
            isinstance(header, Iterable) and not isinstance(header, str)
            for header in headers
        ]
        if all(headers_iterables):
            return True
        # if we have a mixture of iterables and non-iterables, throw an error
        if any(headers_iterables):
            raise ValueError(
                f"Header list `{headers}` contains both iterables and non-iterables."
            )
        return False

    def add_row(
        self,
        title: Optional[str],
        columns: List[Union[str, int, float]]
    ) -> tr:
        """Add a row of cells to the table."""
        with self.body:
            with tr() as row:
                if title is not None:
                    th(title, scope="row1")
                for column in columns:
                    td(raw(str(column)))
        return row

    def _export_table(self, datatable_render) -> str:
        """Add a button for exporting the datatable."""
        # Add button
        exportCSVButton = f"{self.uid}_exportButton"
        button(
            "Export CSV", id=exportCSVButton, type="button",
            className=cls("btn btn-outline-primary"))
        datatable_render = datatable_render + \
            """
            document.getElementById('{{ button_id }}').addEventListener('click',\
            () => {{ id }}_table.export({ \
            type: 'csv', download: true, filename: '{{file_name}}'}))
            """
        return datatable_render, exportCSVButton

    @classmethod
    def from_pandas(
        cls,
        df: pd.DataFrame,
        use_index: bool = True,
        **kwargs
    ):
        """
        Create a `DataTable` from a `pandas.DataFrame`.

        Multi-level columns are supported but multi-level indices are currently not.

        :param df: `pandas.DataFrame`
        :param use_index: Include `DataFrame` index column in table, defaults to `True`
        :return: `DataTable`
        """
        nlevels = df.columns.nlevels
        headers = list(df.columns)

        if nlevels > 1:
            # multiindex columns --> convert into list of tuples
            headers = []
            for i in range(nlevels):
                col_lvl = df.columns.get_level_values(i).to_list()
                if use_index:
                    if i == nlevels - 1:
                        # Set the index name in the bottom level of the table columns.
                        col_lvl.insert(0, df.index.name or 'index')
                    else:
                        col_lvl.insert(0, '')
                headers.append(col_lvl)

        else:
            if use_index:
                headers = [df.index.name or "index", *headers]
        dtable = cls(headers=headers, **kwargs)

        for idx, *row in df.itertuples():
            title = idx if use_index else None
            dtable.add_row(title=title, columns=row)

        return dtable

    @classmethod
    def from_dict(cls, d: dict, use_index: bool = True, **kwargs):
        """Create `DataTable` from a dictionary.

        Uses `pd.DataFrame.from_dict()` and creates the DataTable from the resulting
        `pd.DataFrame`. Be aware that `pd.DataFrame.from_dict()` by default assumes
        "column orientation" (i.e. each key--value pair denotes a column name and
        collection of column values). You can change this behaviour with
        `orient="index"`.

        :param d: Input dictionary; values should be list/tuple-like collections but not
            other dictionaries. Scalar values will cause an error in
            `pd.DataFrame.from_dict()`.
        :param use_index: Include index column of generated DataFrame in table, defaults
            to `True`
        :return: `DataTable`
        """
        df = pd.DataFrame.from_dict(d, **kwargs)
        return cls.from_pandas(df, use_index=use_index)



"""Utility functions for aiding plotting."""
from itertools import cycle, islice
import os

import numpy as np
import pandas as pd
import pkg_resources
from scipy import stats as sp_stats

from ezcharts import plots, util

sns_type_to_echarts = {
    "categorical": "category",
    "numeric": "value"}


class JSCode():
    """A class for javascript code snippets."""

    DELIMITER = "~~~0^0~~~"

    def __init__(self, jscode):
        """Initialise object."""
        self.jscode = jscode

    def to_json(self):
        """Output json with some stuff so we can strip them out."""
        return f"{self.DELIMITER}{self.jscode}{self.DELIMITER}"

    @classmethod
    def _clean(cls, jscode):
        """
        Clean up escape characters in the code.

        This method is intended for use after JSON serialisation.
        """
        dirty = [
            '"%s' % cls.DELIMITER,
            '%s"' % cls.DELIMITER,
            "\\n"]

        for find in dirty:
            jscode = jscode.replace(find, '')

        return jscode


def read_files(summaries, sep="\t"):
    """Read a set of files and join to single dataframe."""
    dfs = list()
    if not isinstance(summaries, (list, tuple)):
        summaries = [summaries]
    for fname in sorted(summaries):
        dfs.append(pd.read_csv(fname, sep=sep))
    return pd.concat(dfs)


def concat_dfs_with_categorical_columns(dfs):
    """Concatenate collection of dataframes while maintaining categorical dtypes.

    :param dfs: Collection of input dataframes
    :raises ValueError: Raised when columns in all the dataframes don't match
    :return: concatenated dataframe with categorical columns

    `pd.concat` will revert the `dtype` of categorical columns to `object` if the
    columns don't contain the same categories in all dataframes. We use
    `union_categoricals` to avoid this.

    Note: This will also change the categories of the categorical dtypes in the input
    dataframes.
    """
    # if there is only one df in the list, return it
    if len(dfs) == 1:
        return dfs[0]
    # make sure all dfs contain the same columns and dtypes
    cols_dtypes = {col: dtype.name for col, dtype in dfs[0].dtypes.items()}
    for df in dfs[1:]:
        if cols_dtypes != {col: dtype.name for col, dtype in df.dtypes.items()}:
            raise ValueError("All DataFrames need to have the same columns and dtypes.")
    # find the categorical columns
    cat_cols = []
    non_cat_cols = []
    for colname, coltype in dfs[0].dtypes.items():
        if isinstance(coltype, pd.api.types.CategoricalDtype):
            cat_cols.append(colname)
        else:
            non_cat_cols.append(colname)
    # concat the non-categorical columns first
    res_df = pd.concat((df[non_cat_cols]) for df in dfs)
    # separately concat the categorical columns in a way that preserves the categories
    # and add them to the results df
    for col in cat_cols:
        res_df[col] = pd.api.types.union_categoricals(
            [df[col] for df in dfs], sort_categories=True
        )
    # make sure that the dtypes of the concatenated df are the same as the input dfs
    if cols_dtypes != {col: dtype.name for col, dtype in res_df.dtypes.items()}:
        raise ValueError(
            "Concatenated DataFrame has different dtypes from input DataFrames."
        )
    # return (with the initial column order)
    return res_df[dfs[0].columns].reset_index(drop=True)


class _colors:
    """Some colours that someone thought were nice."""

    cerulean = "#0084A9"
    not_black = "#001A21"
    feldgrau = "#455556"
    dim_gray = "#666666"
    light_cornflower_blue = "#90C5E7"
    dark_gray = "#B5AEA7"
    isabelline = "#F0EFED"
    medium_spring_bud = "#B8E986"
    cinnabar = "#EF4134"
    sandstorm = "#F5CC49"
    fandango = "#A53F96"
    green = "#17BB75"
    verdigris = "#54B8B1"

    white = "#FFFFFF"
    black = "#000000"
    grey10 = "#1A1A1A"
    grey20 = "#333333"
    grey30 = "#4D4D4D"
    grey40 = "#666666"
    grey50 = "#808080"
    grey60 = "#999999"
    grey70 = "#B3B3B3"
    grey80 = "#CCCCCC"
    grey90 = "#E6E6E6"


class _ontColors:
    """ONT Brand colors."""

    BRAND_BLUE = '#0084A9'
    BRAND_BLUE_LIGHT = '#0084a91a'
    BRAND_LIGHT_BLUE = '#90C6E7'
    BRAND_BLACK = '#333333'
    BRAND_LIGHT_GREY = '#B5AEA7'
    BRAND_GREY = '#666666'
    BRAND_GREY_LIGHT = '#6666661f'

    BRAND_LOGO = pkg_resources.resource_filename(
        'ezcharts', 'data/images/ONT_logo.txt')
    with open(BRAND_LOGO, 'r', encoding="UTF-8") as fh:
        BRAND_LOGO = fh.read()


class _ondColors:
    """OND Brand colors."""

    BRAND_BLUE = '#003E71'
    BRAND_BLUE_LIGHT = '#003e7136'
    BRAND_LIGHT_BLUE = '#90C6E7'
    BRAND_BLACK = '#333333'
    BRAND_LIGHT_GREY = '#B5AEA7'
    BRAND_GREY = '#666666'
    BRAND_GREY_LIGHT = '#6666661f'
    BRAND_RED = '#f45b69'

    BRAND_LOGO = pkg_resources.resource_filename(
        'ezcharts', 'data/images/OND_logo.txt')
    with open(BRAND_LOGO, 'r', encoding="UTF-8") as fh:
        BRAND_LOGO = fh.read()


Colors = _colors()
ont_colors = _ontColors()
ond_colors = _ondColors()


def kernel_density_estimate(x, step=0.2):
    """Kernel density to approximate distribution.

    :param x: data of which to find mode.
    :param step: discretization of KDE PDF.
    """
    # estimate bandwidth of kde, R's nrd0 rule-of-thumb
    hi = np.std(x, ddof=1)
    q75, q25 = np.percentile(x, [75, 25])
    iqr = q75 - q25
    lo = min(hi, iqr/1.34)
    if not ((lo == hi) or (lo == abs(x[0])) or (lo == 1)):
        lo = 1
    bw = 0.9 * lo * len(x)**-0.2

    # create a KDE
    x_grid = np.arange(min(x), max(x), step)
    kernel = sp_stats.gaussian_kde(x, bw_method=bw)
    pdf = kernel(x_grid)
    return x_grid, pdf


def choose_palette(name='colorblind', ncolours=None):
    """Choose colour palette.

    These are seaborn's qualitative color palettes.
    https://seaborn.pydata.org/tutorial/color_palettes.html

    :param: name: name of palette to return colors from.
    :param: ncolours: number of colours.
    """
    palettes = {
        'colorblind': [
            '#0173b2', '#de8f05', '#029e73', '#d55e00', '#cc78bc',
            '#ca9161', '#fbafe4', '#949494', '#ece133', '#56b4e9'],
        'deep': [
            '#4c72b0', '#dd8452', '#55a868', '#c44e52', '#8172b3',
            '#937860', '#da8bc3', '#8c8c8c', '#ccb974', '#64b5cd'],
        'dark': [
            '#001c7f', '#b1400d', '#12711c', '#8c0800', '#591e71',
            '#592f0d', '#a23582', '#3c3c3c', '#b8850a', '#006374'],
        'muted': [
            '#4878d0', '#ee854a', '#6acc64', '#d65f5f', '#956cb4',
            '#8c613c', '#dc7ec0', '#797979', '#d5bb67', '#82c6e2'],
        'bright': [
            '#023eff', '#ff7c00', '#1ac938', '#e8000b', '#8b2be2',
            '#9f4800', '#f14cc1', '#a3a3a3', '#ffc400', '#00d7ff'],
        'pastel': [
            '#a1c9f4', '#ffb482', '#8de5a1', '#ff9f9b', '#d0bbff',
            '#debb9b', '#fab0e4', '#cfcfcf', '#fffea3', '#b9f2f0'],
        # Uniform palettes
        'rocket': [
            '#35193e', '#701f57', '#ad1759', '#e13342', '#f37651',
            '#f6b48f'],
        'mako': [
            '#2e1e3b', '#413d7b', '#37659e', '#348fa7', '#40b7ad',
            '#8bdab2'],
        'flare': [
            '#e98d6b', '#e3685c', '#d14a61', '#b13c6c', '#8f3371',
            '#6c2b6d'],
        'crest': [
            '#7dba91', '#59a590', '#40908e', '#287a8c', '#1c6488',
            '#254b7f'],
        'magma': [
            '#221150', '#5f187f', '#982d80', '#d3436e', '#f8765c',
            '#febb81'],
        'viridis': [
            '#46327e', '#365c8d', '#277f8e', '#1fa187', '#4ac16d',
            '#a0da39'],
        'rocket_r': [
            '#f6b48f', '#f37651', '#e13342', '#ad1759', '#701f57',
            '#35193e'],
        # Uniform diverging palettes
        'vlag': [
            '#6e90bf', '#aab8d0', '#e4e5eb', '#f2dfdd', '#d9a6a4',
            '#c26f6d'],
        'icefire': [
            '#55a3cd', '#4954b0', '#282739', '#3b2127', '#9c2f45',
            '#e96f36']
    }

    if ncolours is None:
        return palettes[name]

    cycler = cycle(palettes[name])
    return [c for c in islice(cycler, ncolours)]


def empty_plot(**kwargs):
    """Empty plot for when all else fails.

    :param kwargs: kwargs for bokeh figure.

    :returns: an eChart empty plot.

    """
    plt = plots._NoAxisFixPlot()
    plt.title = {
        "text": "Plotting Failed",
        "subtext": "Oops! Something went wrong in plotting your data.",
        "subtextStyle": {
            "fontStyle": "italic",
            "color": Colors.cerulean,
        },
        "itemGap": 30,
    }
    return plt


def plot_wrapper(func):
    """Decorate plotting functions to ignore exceptions.

    :param func: plotting function.
    :param args: the arguments provided.
    :param kwargs: other optional key word arguments.

    :returns: plot.

    """
    def wrapper_accepting_arguments(*args, **kwargs):
        """Argument wrapper for decorator."""
        logger = util.get_named_logger("PlotWrap")
        try:
            p = func(*args, **kwargs)
            return p
        except Exception as e:
            # check if debug mode is enabled:
            if os.environ.get("EZCHARTS_DEBUG") == "1":
                raise e
            else:
                # plot empty
                try:
                    p = empty_plot(**kwargs)
                    logger.exception("Plotting user plot failed with: " + str(e))
                    return p
                # if empty plot fails
                except Exception as f:
                    p = empty_plot()
                    logger.exception(
                        "Plotting templated Empty plot failed with: " + str(f)
                    )
                    return p

    return wrapper_accepting_arguments


class Limiter(object):
    """Helper class to determine limits of data."""

    def __init__(self, limits=None):
        """Initialize the limit gathering."""
        if limits is not None:
            self.min, self.max = limits
        else:
            self.min, self.max = +float('inf'), -float('inf')

    def accumulate(self, data):
        """Analyse data to update limits.

        :param data: new data to analyse.

        """
        if len(data) > 0:
            self.min = min(self.min, min(data))
            self.max = max(self.max, max(data))
        return self

    def fix(self, lower=None, upper=None):
        """Fix limits to new values.

        :param lower: new lower limit.
        :param upper: new upper limit.

        """
        if lower is not None:
            self.min = lower
        if upper is not None:
            self.max = upper
        return self

    def __repr__(self):
        """Return string representation of self."""
        return repr((self.min, self.max))



"""Visualizing matrices of data."""
import pandas as pd
import seaborn as sns

from ezcharts.plots import Plot


__all__ = ["heatmap", "clustermap"]


def heatmap(
        data, *, vmin=None, vmax=None, cmap=None, center=None, robust=False,
        annot=True, fmt='.2g', annot_kws=None, linewidths=0, linecolor='white',
        cbar=True, cbar_kws=None, cbar_ax=None, square=False, xticklabels='auto',
        yticklabels='auto', mask=None, ax=None, **kwargs):
    """
    Plot rectangular data as a color-encoded matrix.

    Accepts a 2D dataset (list of lists) that can be coerced into an ndarray,
    or a Pandas dataframe.

    Data should be in a matrix where columns are the heatmap x axis and rows
    are the heatmap y axis.

    Axes titles are the name of the index or the columns. You can rename them
    like so:
    - index (x axis)
    data = data.rename_axis("x_axis_name")
    - columns (y axis)
    data = data.rename_axis("y_axis_name", axis=1)
    """
    # 2D arrays need to be reshaped using melt so they are the correct format
    # for add_dataset
    if not isinstance(data, pd.DataFrame):
        data = pd.DataFrame(data)

    # Get data in format for ezcharts
    data = data.melt(ignore_index=False).reset_index().round(1)

    # If the user doesn't provide colors then default to something
    # that looks nice in our reports
    if cmap is None:
        cmap = sns.color_palette("viridis",  n_colors=1000, as_cmap=False).as_hex()

    # If the user doesn't provide minimum and maximums then work these
    # out from the data
    if vmin is None:
        vmin = data['value'].min()

    if vmax is None:
        vmax = data['value'].max()

    # Get the names for the axes
    x_name, y_name, plot_data = data.columns

    plt = Plot()

    plt.xAxis = dict(
        name=x_name,
        type='category',
        axisLabel=dict(
            interval=0,
            rotate=40
        ),
        axisTick=dict(
            alignWithLabel=True))

    plt.yAxis = dict(
        name=y_name,
        type='category',
        axisTick=dict(
            alignWithLabel=True))

    # On mouse over emphasize the heatmap tile
    emphasis_style = dict(
        itemStyle=dict(
            shadowBlur=10,
            shadowColor='#666666'))

    plt.series = [dict(
        type='heatmap',
        label=dict(show=annot),
        itemStyle=dict(borderWidth=linewidths, borderColor=linecolor),
        emphasis=emphasis_style)]

    plt.visualMap = [dict(
        min=vmin,
        max=vmax,
        calculable=True,
        left='right',
        top='center',
        orient='vertical',
        inRange=dict(color=cmap))]

    plt.add_dataset(dict(
        id='raw',
        dimensions=data.columns,
        source=data.values))

    return plt


def clustermap(*args, **kwargs):
    """Plot a matrix dataset as a hierarchically-clustered heatmap."""
    raise NotImplementedError



"""Get default section layouts."""
from typing import Optional, Type

from dominate.tags import h2, html_tag

from ezcharts.layout.base import IClasses, IStyles, Snippet
from ezcharts.layout.util import cls, css


class ISectionClasses(IClasses):
    """Section html classes."""

    container: str = cls(
        "container", "p-4", "mb-4",
        "bg-white", "border", "rounded")
    container_trans: str = cls("container", "py-1", "mb-4")
    title: str = cls("h5", "mb-0", "pb-3")


class ISectionStyles(IStyles):
    """Section html classes."""

    overflow: str = css("overflow-x: auto")


class Section(Snippet):
    """A styled section snippet."""

    TAG = "section"

    def __init__(
        self,
        section_id: str,
        section_title: Optional[str] = None,
        section_title_tag: Type[html_tag] = h2,
        styles: ISectionStyles = ISectionStyles(),
        classes: ISectionClasses = ISectionClasses(),
        overflow: bool = False,
        transparent: bool = False
    ) -> None:
        """Create styled section."""
        super().__init__(
            styles=styles,
            classes=classes,
            className=classes.container_trans if transparent else classes.container,
            style=styles.overflow if overflow else None,
            id=section_id)

        if section_title and section_title_tag:
            with self:
                section_title_tag(
                    section_title,
                    className=self.classes.title)
